class AopError(Exception):
    """
    1. Failed before sending a request.
        e.g. Some of the required API parameters missing.
    2. Failed to parse the returned results.

    """
    pass


class ApiError(Exception):
    """
    The remote server returned error and the error messages were successfully recognized.

    Attributes
    ----------
    api : str
    error_code : str
    error_message : str
    exception : str
    request_id : str

    """

    def __init__(self, api, error_code='', error_message='', exception='', request_id=''):
        self.api = api
        self.error_code = error_code
        self.error_message = error_message
        self.exception = exception
        self.request_id = request_id

    def __str__(self, *args, **kwargs):
        return 'ApiError [api=%s; error_code=%s; error_message=%s; exception=%s; request_id=%s]' % \
               (self.api, self.error_code, self.error_message, self.exception, self.request_id)



# -*- coding: utf-8 -*-
from aop.api.base import BaseApi


class AlibabaTradeGetBuyerViewParam(BaseApi):
    """获取单个交易明细信息，仅限买家调用。该API需要向阿里巴巴开放平台申请权限才能使用。Get a single transaction detail, only for users to call.

    References
    ----------
    https://open.1688.com/api/api.htm?ns=com.alibaba.trade&n=alibaba.trade.get.buyerView&v=1&cat=order_category

    """

    def __init__(self, app_info, access_token, order_id, *, domain=None, web_site='webSite', include_fields=None,
                 attribute_keys=None):
        BaseApi.__init__(self, app_info, domain)
        self.access_token = access_token
        self.webSite = web_site
        self.orderId = order_id
        self.includeFields = include_fields
        self.attributeKeys = attribute_keys

    def get_api_uri(self):
        return '1/com.alibaba.trade/alibaba.trade.get.buyerView'

    def get_required_params(self):
        return ['webSite', 'orderId']

    def get_multipart_params(self):
        return []

    def need_sign(self):
        return True

    def need_timestamp(self):
        return False

    def need_auth(self):
        return True

    def need_https(self):
        return False

    def is_inner_api(self):
        return False



# -*- coding: utf-8 -*-
from aop.api.base import BaseApi


class AlibabaTradeGetBuyerOrderListParam(BaseApi):
    """获取买家的订单列表，也就是用户的memberId必须等于订单里的买家memberId。该接口仅仅返回订单基本信息，不会返回订单的物流信息和发票信息；如果需要获取物流信息，请调用获取订单详情接口；如果需要获取发票信息，请调用获取发票信息的API

    References
    ----------
    https://open.1688.com/api/api.htm?ns=com.alibaba.trade&n=alibaba.trade.getBuyerOrderList&v=1&cat=order_category

    """

    def __init__(self, app_info, access_token, domain=None, *, biz_types=None, create_end_time=None,
                 create_start_time=None, is_his=None, modify_end_time=None, modify_start_time=None, order_status=None,
                 page=None, page_size=None, refund_status=None, seller_member_id=None, seller_login_id=None,
                 seller_rate_status=None, trade_type=None, product_name=None, need_buyer_address_and_phone=None,
                 need_memo_info=None):
        BaseApi.__init__(self, app_info, domain)
        self.access_token = access_token
        self.bizTypes = biz_types
        self.createEndTime = create_end_time
        self.createStartTime = create_start_time
        self.isHis = is_his
        self.modifyEndTime = modify_end_time
        self.modifyStartTime = modify_start_time
        self.orderStatus = order_status
        self.page = page
        self.pageSize = page_size
        self.refundStatus = refund_status
        self.sellerMemberId = seller_member_id
        self.sellerLoginId = seller_login_id
        self.sellerRateStatus = seller_rate_status
        self.tradeType = trade_type
        self.productName = product_name
        self.needBuyerAddressAndPhone = need_buyer_address_and_phone
        self.needMemoInfo = need_memo_info

    def get_api_uri(self):
        return '1/com.alibaba.trade/alibaba.trade.getBuyerOrderList'

    def get_required_params(self):
        return []

    def get_multipart_params(self):
        return []

    def need_sign(self):
        return True

    def need_timestamp(self):
        return False

    def need_auth(self):
        return True

    def need_https(self):
        return False

    def is_inner_api(self):
        return False



# -*- coding: utf-8 -*-
from aop.api.base import BaseApi
from aop.structure import Address, Cargo, Invoice


class AlibabaTradeFastCreateOrderParam(BaseApi):
    """快速创建1688大市场订单和1688代销订单，订单一步创建，不需要先调用订单预览，接口参数简单，地址参数传省市区街道的文本名，不需要额外查询地址码，系统默认选择最优惠下单方式，默认支付宝担保交易方式，详细地址必须不超过200个字，不要用地址做其他用途，需要留言或备注的有专门字段，留言和备注都支持500字

    References
    ----------
    https://open.1688.com/api/api.htm?ns=com.alibaba.trade&n=alibaba.trade.fastCreateOrder&v=1&cat=aop.trade

    """

    def __init__(self, app_info, access_token, address: Address, cargo: Cargo, *, flow="general", domain=None,
                 sub_user_id=None, message="", invoice: Invoice = None, shop_promotion_id="", trade_type=""):
        BaseApi.__init__(self, app_info, domain)
        self.access_token = access_token
        self.flow = flow
        self.subUserId = sub_user_id
        self.message = message
        self.addressParam = address.get_data()
        self.cargoParamList = cargo.get_data()
        self.invoiceParam = invoice.get_data() if invoice else None
        self.shopPromotionId = shop_promotion_id
        self.tradeType = trade_type

    def get_api_uri(self):
        return '1/com.alibaba.trade/alibaba.trade.fastCreateOrder'

    def get_required_params(self):
        return ['flow', 'addressParam', 'cargoParamList']

    def get_multipart_params(self):
        return []

    def need_sign(self):
        return True

    def need_timestamp(self):
        return False

    def need_auth(self):
        return True

    def need_https(self):
        return False

    def is_inner_api(self):
        return False



from aop.api.base import FileItem, BaseApi
from aop.api.transaction.trade_receive_address import AlibabaTradeReceiveAddressGetParam
from aop.api.transaction.trade_buyer_order_list import AlibabaTradeGetBuyerOrderListParam
from aop.api.transaction.trade_buyer_view import AlibabaTradeGetBuyerViewParam
from aop.api.transaction.trade_fast_create_order import AlibabaTradeFastCreateOrderParam

from aop.api.logistics.logistics_trace_info import AlibabaTradeGetLogisticsTraceInfoBuyerViewParam
from aop.api.logistics.logistic_company import AlibabaLogisticsOpQueryLogisticCompanyListParam



import json


class AppInfo(object):
    def __init__(self, app_key, secret):
        self.app_key = app_key
        self.secret = secret


class Address(object):
    def __init__(self, address_id, full_name, mobile, phone, post_code, city, province, area, town, address,
                 district_code):
        self.addressId = address_id
        self.fullName = full_name
        self.mobile = mobile
        self.phone = phone
        self.postCode = post_code
        self.cityText = city
        self.provinceText = province
        self.areaText = area
        self.townText = town
        self.address = address
        self.districtCode = district_code

    def get_data(self):
        data = {}
        for key, value in self.__dict__.items():
            data[key] = value
        return json.dumps(data)


class Cargo(object):
    def __init__(self):
        self.param_list = []

    def add(self, offer_id, spec_id, quantity):
        data = {
            "offerId": offer_id,
            "specId": spec_id,
            "quantity": quantity,
        }
        self.param_list.append(data)

    def get_data(self):
        return json.dumps(self.param_list)


class Invoice(object):
    def __init__(self, invoice_type, province, city, area, town, address, post_code, district_code, full_name, phone,
                 mobile, company_name, tax_payer_identifier, bank_account, local_invoice_id):
        self.invoiceType = invoice_type
        self.fullName = full_name
        self.mobile = mobile
        self.phone = phone
        self.postCode = post_code
        self.cityText = city
        self.provinceText = province
        self.areaText = area
        self.townText = town
        self.address = address
        self.districtCode = district_code
        self.companyName = company_name
        self.taxpayerIdentifier = tax_payer_identifier
        self.bankAndAccount = bank_account
        self.localInvoiceId = local_invoice_id

    def get_data(self):
        data = {}
        for key, value in self.__dict__.items():
            data[key] = value
        return json.dumps(data)



# -*- coding: utf-8 -*-
from aop.api.base import BaseApi


class AlibabaTradeReceiveAddressGetParam(BaseApi):
    """买家获取保存的收货地址信息列表

    References
    ----------
    https://open.1688.com/api/api.htm?ns=com.alibaba.trade&n=alibaba.trade.receiveAddress.get&v=1&cat=aop.trade

    """

    def __init__(self, app_info, access_token, domain=None):
        super().__init__(app_info, domain)
        self.access_token = access_token

    def get_api_uri(self):
        return '1/com.alibaba.trade/alibaba.trade.receiveAddress.get'

    def get_required_params(self):
        return []

    def get_multipart_params(self):
        return []

    def need_sign(self):
        return True

    def need_timestamp(self):
        return False

    def need_auth(self):
        return True

    def need_https(self):
        return False

    def is_inner_api(self):
        return False



import time
import hmac
import hashlib
import requests
import json
from aop.structure import AppInfo
from aop.exceptions import AopError, ApiError
from abc import abstractmethod, ABCMeta
from requests.exceptions import HTTPError

# constants
P_OPENAPI = 'openapi'
P_API = 'api'
P_PARAM2 = 'param2'

# system parameters
P_ACCESS_TOKEN = 'access_token'
P_TIMESTAMP = '_aop_timestamp'
P_SIGN = '_aop_signature'

P_ERROR_CODE = 'error_code'
P_ERROR_MESSAGE = 'error_message'
P_EXCEPTION = 'exception'
P_REQUEST_ID = 'request_id'

P_SYS_PARAMS = [P_TIMESTAMP, P_SIGN, P_ACCESS_TOKEN]


class FileItem(metaclass=ABCMeta):
    def __init__(self, filename=None, content=None):
        """
        Parameters
        ----------
        content : file-like-object(str/bytes/bytearray or an object that has read attribute)
            e.g. f = open('filepath', 'rb') ... fileitem = aop.api.FileItem('filename', f)
            e.g. fileitem = aop.api.FileItem('filename', 'file_content_str')

        """
        self.filename = filename
        self.content = content


class BaseApi(object):
    """"Base class of APIs.

    """

    def __init__(self, app_info: AppInfo, domain=None):
        """
        初始化接口
        :param app_info:
        :param domain:
        """
        self.__domain = domain or "gw.open.1688.com"
        self.__http_method = "POST"
        self.__app_key = app_info.app_key
        self.__secret = app_info.secret

    def raise_aop_error(self, *args):
        _args = ['API: ' + self.get_api_uri()]
        if args:
            _args.extend(args)
        raise AopError(*_args)

    def sign(self, url_path, params, secret):
        """
        Method to generate _aop_signature.
        Parameters
        ----------
        url_path : str param2/version/namespace/name/appkey
        params : dict-like
        secret : str
        References
        ----------
        https://open.1688.com/api/sysSignature.htm

        """
        if not url_path:
            self.raise_aop_error('sign error: urlPath missing')
        if not secret:
            self.raise_aop_error('sign error: secret missing')

        param_list = []
        if params:
            if not hasattr(params, "items"):
                self.raise_aop_error('sign error: params must be dict-like')
            param_list = [k + v for k, v in params.items()]
            param_list = sorted(param_list)

        msg = bytearray(url_path.encode('utf-8'))
        for param in param_list:
            msg.extend(bytes(param.encode('utf-8')))

        sha = hmac.new(bytes(secret.encode('utf-8')), None, hashlib.sha1)
        sha.update(msg)
        return sha.hexdigest().upper()

    @abstractmethod
    def get_api_uri(self):
        pass

    def need_sign(self):
        """True if _aop_signature is needed"""
        return False

    def need_timestamp(self):
        """True if _aop_timestamp is needed"""
        return False

    def need_auth(self):
        """True if access_token is needed"""
        return False

    def need_https(self):
        """True if to send a https request"""
        return False

    def is_inner_api(self):
        """True if not an open api. Usually false."""
        return False

    def get_multipart_params(self):
        """
        Returns
        -------
        list
            API parameters with type of byte[].
            A multipart request will be sent if not empty.
            e.g. returns ['image'] if the 'image' parameter's type is byte[].
                Then assign it as req.image = aop.api.FileItem('imagename.png', imagecontent)

        """
        return []

    def get_required_params(self):
        """
        Names of required API parameters.
        System parameters(access_token/_aop_timestamp/_aop_signature) excluded.

        An AopError with message 'Required params missing: {"missing_param1_name", ...}'
        will be thrown out if some of the required API parameters are missing before a
        request sent to the remote server.

        """
        return []

    def __get_url_protocol(self):
        return 'https' if self.need_https() else 'http'

    def _build_sign_url_path(self):
        return '%s/%s/%s' % (P_PARAM2, self.get_api_uri(), self.__app_key)

    def _build_url(self, sign_url_path):
        return '%s://%s/%s/%s' % (self.__get_url_protocol(), self.__domain,
                                  P_API if self.is_inner_api() else P_OPENAPI, sign_url_path)

    def _check_sign(self):
        if self.need_sign():
            if not self.__app_key:
                self.raise_aop_error('AppKey missing')
            if not self.__secret:
                self.raise_aop_error('App secret missing')

    def _check_auth(self, **kwargs):
        if self.need_auth():
            if not kwargs.get(P_ACCESS_TOKEN):
                self.raise_aop_error('access_token missing')

    def _check_required_params(self, **kwargs):
        missing_params = set(self.get_required_params()) - set(kwargs.keys())
        if missing_params:
            self.raise_aop_error('Required params missing: %s' % (str(missing_params)))

    @staticmethod
    def _gen_timestamp():
        return int(time.time() * 1000)

    def get_response(self, timeout=5, **kwargs):

        params = self._get_nonnull_biz_params()
        if kwargs:
            params.update(kwargs)

        if self.need_sign():
            self._check_sign()
        if self.need_auth():
            self._check_auth(**params)
        if self.get_required_params():
            self._check_required_params(**params)
        if self.need_timestamp() and (not params.get(P_TIMESTAMP)):
            params[P_TIMESTAMP] = self._gen_timestamp(self.__domain)

        for multipart_param in self.get_multipart_params():
            params.pop(multipart_param)

        sign_url_path = self._build_sign_url_path()
        if self.need_sign() and (not params.get(P_SIGN)):
            params[P_SIGN] = self.sign(sign_url_path, params, self.__secret)

        url = self._build_url(sign_url_path)
        headers = self._get_request_header();

        files = {}
        if self.get_multipart_params():
            for key in self.get_multipart_params():
                fileitem = getattr(self, key)
                if (fileitem and isinstance(fileitem, FileItem)):
                    files[key] = (fileitem.filename, fileitem.content)
                else:
                    self.raise_aop_error(key + ' not a FileItem')
        response = self._do_request(url, params, files, headers=headers, timeout=timeout)

        return response

    def _do_request(self, url, data=None, files=None, **kwargs):
        try:
            if files:
                response = requests.post(url, data=data, files=files, **kwargs)
            else:
                response = requests.post(url, data=data, **kwargs)

            response.raise_for_status()
            return response.json()

        except HTTPError as e:
            if not (response.status_code >= 200 and response.status_code < 400):
                response_json = response.json()
                param = {
                    "api": self.get_api_uri(),
                    "error_code": response_json.get(P_ERROR_CODE),
                    'error_message': response_json.get(P_ERROR_MESSAGE),
                    "exception": response_json.get(P_EXCEPTION),
                    'request_id': response_json.get(P_REQUEST_ID)
                }
                raise ApiError(**param)

            raise e
        except requests.RequestException as e:
            self.raise_aop_error(str(e))

    @staticmethod
    def _get_request_header():
        return {
            'Cache-Control': 'no-cache',
            'Connection': 'Keep-Alive',
            'User-Agent': 'Ocean-SDK-Client'
        }

    @staticmethod
    def _is_nonnull_biz_param(param_name, param_value):
        if param_name in P_SYS_PARAMS:
            return True
        return (not param_name.startswith("__")) and (not param_name.startswith("_BaseApi__")) and param_value

    def _get_nonnull_biz_params(self):
        biz_params = {}
        for key, value in self.__dict__.items():
            if self._is_nonnull_biz_param(key, value):
                biz_params[key] = value
        return biz_params



# -*- coding: utf-8 -*-
from aop.api.base import BaseApi


class AlibabaTradeGetLogisticsTraceInfoBuyerViewParam(BaseApi):
    """该接口需要获取订单买家的授权，获取买家的订单的物流跟踪信息，在采购或者分销场景中，作为买家也有获取物流详情的需求。该接口能查能根据物流单号查看物流单跟踪信息。由于物流单录入的原因，可能跟踪信息的API查询会有延迟。该API需要向开放平台申请权限才能访问。In the procurement or distribution scenario, buyers can obtain information on logistics tracking. The interface can view the logistics tracking information according to the logistics tacking number. Depending on the logistics information entry time, there may be a delay in API queries regarding the information tracking.

    References
    ----------
    https://open.1688.com/api/api.htm?ns=com.alibaba.logistics&n=alibaba.trade.getLogisticsTraceInfo.buyerView&v=1&cat=wuliu

    """

    def __init__(self, app_info, access_token, order_id, *, domain=None, web_site='1688', logistics_id=''):
        BaseApi.__init__(self, app_info, domain)
        self.access_token = access_token
        self.logisticsId = logistics_id
        self.orderId = order_id
        self.webSite = web_site

    def get_api_uri(self):
        return '1/com.alibaba.logistics/alibaba.trade.getLogisticsTraceInfo.buyerView'

    def get_required_params(self):
        return ['orderId', 'webSite']

    def get_multipart_params(self):
        return []

    def need_sign(self):
        return True

    def need_timestamp(self):
        return False

    def need_auth(self):
        return True

    def need_https(self):
        return False

    def is_inner_api(self):
        return False
